## @file CurveT.py
#  @author Steven Palmer and Henry Madej
#  @brief CurveT
#  @date 20/02/2018

from Exceptions import *
from SeqServices import *




## @brief An abstract data type that represents a Curve
class CurveT:

  MAX_ORDER = 2
  DX = 1E-3

  ## @brief CurveT constructor
  #  @details takes two lists of of x's and y's where the values are sorted in 
  #  ascending order, and order i of the function representing the curve
  #  @param X list of x values sorted in ascending order
  #  @param Y list of corresponding y values to the the x's in the X list
  #  @param i order of the function represented by the data points (x,y) pairs
  def __init__(self, X, Y, i):
    if not isAscending(X):
      raise IndepVarNotAscending
    if i < 1 or i > CurveT.MAX_ORDER:
      raise InvalidInterpOrder
    if len(X) != len(Y):
      raise SeqSizeMismatch
      
    self.__minx = X[0] 
    self.__maxx = X[-1]
    self.__o = i
    
    if self.__o == 1:
      self.__f = lambda v: interpLin(X[index(X, v)], Y[index(X, v)], X[index(X, v) + 1], Y[index(X, v) + 1], v)  
    elif self.__o == 2:
      self.__f = lambda v: interpQuad(X[index(X, v) - 1], Y[index(X, v) - 1], X[index(X, v)], Y[index(X, v)], X[index(X, v) + 1], Y[index(X, v) + 1], v)
  
  ## @brief minD returns the minimal x value
  #  @return value representing the minimal x value
  def minD(self):
    return self.__minx
  
  ## @brief maxD returns the maximal x value
  #  @return value representing the maximal x value
  def maxD(self):
    return self.__maxx
    

  ## @brief order returns the order of the function represented by the curve
  #  @return the value of the order of the curve
  def order(self):
    return self.__o
  

  ## @brief eval evaluates the function at a given x value
  #  @details if x is within the range of x values for the curve it will return
  #           a y value for the curve
  #  @param x value of x to evaluate the curve at
  #  @exception OutOfDomain - if x value is not within the range of the curve's 
  #             x values
  #  @return the value of the curve at x
  def eval(self, x):
    if x < self.__minx or x > self.__maxx:
      raise OutOfDomain
    return self.__f(x)


  ## @brief approximates the first derivate of the curve
  #  @details uses forward divided difference
  #  @param x value of x for the first derivate to be evaluated at
  #  @return the value of the first derivate at evaluated x
  def dfdx(self, x):
    if x < self.__minx or x > self.__maxx:
      raise OutOfDomain
    return self.__f(x + self.DX) - self.__f(x) / self.DX

  ## @brief approximates the second derivate of the curve
  #  @details uses forward divided difference
  #  @param x value of x for the second derivate to be evaluated at
  #  @return the value of the second derivate at evaluated x
  def d2fdx2(self, x):
    if x < self.__minx or x > self.__maxx:
      raise OutOfDomain
    return self.__f(x + 2*self.DX) - 2*self.__f(x + self.DX) + self.__f(x) / self.DX**2